///\file main.cpp
///standard winmain using nehe libs and directX wrappers.
///kicks you into a fullscreen opengl ap and swaps buffers until you hit ESC.

#include "main.h"
#include "ui.h"
#include "process_frame.h"

#include <nvidia/lazy_cg.h>

#include <nehe/window.h>
#include <win_input.h>
#include <win_globals.h>
#include <nehe/matrix_stack.h>
#include <sven/gl_array.h>
#include <nehe/matrix_stack.h>
#include <nehe/load_texture.h>
#include <string>
#include <sstream>
#include <iostream>
#include <stdlib.h>
#include <time.h>


#include <reg_helpers.h>

char * app_name = "VisionHW";

bool g_quit_flag=false;

void key_handler() {
	using namespace win_tools;
	if(key_down_event(DIK_ESCAPE)) g_quit_flag=true;
}

void init() {
	glDisable(GL_DITHER);
	glPixelStorei(GL_UNPACK_ALIGNMENT ,1);
	glPixelStorei(GL_PACK_ALIGNMENT ,1);	
	glScreenCoordinateMatrix();

	lazy_cg::init();
	ui::init();
	process_frame::init();
}

void cleanup() {

	ui::cleanup();
		lazy_cg::cleanup();
	process_frame::cleanup();
}


void draw() {
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	process_frame::draw();
	ui::draw();

}

void loop_fun() {
	key_handler();
	ui::update();
	process_frame::update();
	draw();
}


//fairly boilerplate winmain, using a couple of NeHe-based resource wrapping classes
//that i wrote from some of my personal projects
int WINAPI WinMain(HINSTANCE hInst, HINSTANCE hPrevInstance, LPSTR lpszCmdLine, int nCmdShow)
{ 
	MSG msg;
	msg.wParam=-1;
	try {


	srand(time(NULL));

#ifdef MAKE_DISTRO
	set_program_dir(app_name);
	IFL_current_dir();
#endif


	using namespace win_tools;
	using namespace nehe;

	hInstance=hInst;  //Set the global hInstance variable, this will be used all over the place
	clear_input_buffer();

	nehe::Window nehe_win(app_name,screen_w,screen_h,32,win_tools::WI_Proc);

	SetCursor(LoadCursor(NULL, IDC_WAIT));
//	ShowCursor(FALSE);

	DI_ResourceWrapper di_wapper;

	SetCursor(LoadCursor(NULL, IDC_CROSS));	
	init();

	while(!g_quit_flag) {	
		if (PeekMessage(&msg,NULL,0,0,PM_REMOVE)) { 
			if (msg.message == WM_QUIT) break;
			TranslateMessage(&msg);
			DispatchMessage(&msg);
		} 
		else {

			updateDirectInput();

			loop_fun();

			SwapBuffers(hDC);
		}
	}

	cleanup();

	} catch (std::exception &e) {
		MessageBox(NULL,e.what(),"Exception caught in WinMain",MB_OK|MB_ICONEXCLAMATION);
	}

	return (msg.wParam);	
}

