#include <atlcore.h>
#include <atlbase.h>

#include <nvidia/lazy_cg.h>
#include <nvidia/RenderTexture.h>
#include <glh/glh_extensions.h>
#include <gl/glu.h>
#include <sven/gl_array.h>
#include <boost/timer.hpp>
#include <boost/format.hpp>
#include <sstream>

#include <boost/thread/mutex.hpp>
#include <boost/timer.hpp>

#include <sven_ui/commands.h>
#include <sven_ui/template_command.h>
#include <sven/gl_extras.h>

#include <fbitmap.h>
#include <win_globals.h>

namespace process_frame {

using lazy_cg::render_block;
using sven::cstring;

CComAutoCriticalSection  m_cs;
boost::mutex rt_access_mutex;

bool rt_access=false;
RenderTexture * rtexture=NULL;
RenderTexture * rtexture2=NULL;

float * optical_flow = NULL;

int tex_w=1024, tex_h=768;
int source_w=0, source_h=0;

void set_dims(int w, int h) {
	if(optical_flow) delete [] optical_flow;
	
	source_w=w;
	source_h=h;
	optical_flow = new float[source_w*source_h*3];
}

void send_data(unsigned char *data) {

//	if(	source_w!=w ||source_h!=h) throw std::runtime_error(
//			"the source dimensions have been configured improperly.");

	boost::mutex::scoped_lock lock(rt_access_mutex);
	rtexture2->Activate();

	
	glRasterPos2f(0,0);
	glDrawPixels(source_w,source_h,GL_BGR_EXT,GL_UNSIGNED_BYTE,data);
	
	glFlush();
	rtexture2->Deactivate();

}

namespace params {
int r=6;
float sigma=4;
}

void update() {
	if(!optical_flow) return;


	rtexture->Activate();
	
	//step 1, blit the capture data
	{
	
		boost::mutex::scoped_lock lock(rt_access_mutex);
	
		rtexture2->Bind();	
				
		lazy_cg::scoped_fragment bob("rendering","make_grays");	
		render_block(0,0,source_w,source_h,
					0,source_h,source_w,source_h);
		rtexture2->Release();	
		glFlush();
	}
	
	rtexture->Bind();
	
	//step 2, calculate the gradient image
	{		
		lazy_cg::scoped_fragment bob("rendering","make_gradients");
		lazy_cg::set_param("source_h",source_h);
		lazy_cg::set_param("source_w",source_w);
		lazy_cg::set_param("dt",1);
		
		render_block(	0,0,source_w,source_h,
						0,0,source_w,source_h);	
		
		glFlush();	
	}	

	//step 3, calculate the optical flow
	{		
		lazy_cg::scoped_fragment bob("rendering","calc_flow");
		lazy_cg::set_param("source_h",source_h);
		lazy_cg::set_param("source_w",source_w);
		lazy_cg::set_param("r",params::r);
		lazy_cg::set_param("sigma2",params::sigma*params::sigma);
		
		render_block(	0,0,source_w,source_h,
						source_w,0,source_w,source_h);	
		
		
		//we don't need to flush yet, because the copy is
		//independent of the optical flow calc.
		//glFlush();	
	}	
	
	
	//step 4, copy our capture data to the old capture data location
	{
		lazy_cg::scoped_fragment bob("rendering","copy");
		render_block(0,source_h,source_w,source_h,
					source_w,source_h,source_w,source_h);	
			
		glFlush();	
	}
	rtexture->Release();
	
	glReadPixels(source_w,0,source_w,source_h,GL_RGB,GL_FLOAT,optical_flow);
	
	rtexture->Deactivate();
}

const int flow_x=20;
const int flow_y=510;

void draw_flow() {
	if(!optical_flow) return;
	
	glScopedAttribs bob(GL_CURRENT_BIT | GL_ENABLE_BIT);
	glScopedTranslate fred(flow_x,flow_y);
	glDisable(GL_TEXTURE_2D);
	
	const int dx=4;
	const int dy=4;
	glColor3f(1,0,0);
	glPointSize(2.5f);
	glBegin(GL_POINTS);
	//remember that the data on the edges is going to be invalid.
	for(int x=1;x<source_w-1;x+=dx) for(int y=1;y<source_h-1;y+=dy) {
		glVertex2f(x,y);
	}
	glEnd();
	
	glColor3f(1,1,1);
	glBegin(GL_LINES);
	for(int x=0;x<source_w;x+=dx) for(int y=0;y<source_h;y+=dy) {
		glVertex2f(x,y);
		float fx=optical_flow[3*(x+y*source_w)+0];
		float fy=optical_flow[3*(x+y*source_w)+1];
		glVertex2f(x+fx,y+fy);
	}
	glEnd();	
}

void draw() {
	rtexture->Bind();

	glPushAttrib(GL_ENABLE_BIT | GL_CURRENT_BIT);

	int x=10,y=10;

	glPushMatrix();
	glTranslatef(x,y,0);

	glEnable(GL_TEXTURE_2D);
	
	lazy_cg::scoped_fragment bob("rendering","copy");

	glBegin(GL_TRIANGLE_STRIP);	

	glTexCoord2d( tex_w, tex_h);  
	glVertex2f	( tex_w, tex_h);

	glTexCoord2d( 0, tex_h); 
	glVertex2f  ( 0, tex_h);

	glTexCoord2d( tex_w, 0);  
	glVertex2f  ( tex_w, 0);
	
	glTexCoord2d( 0, 0);  
	glVertex2f  ( 0, 0);
	
	glEnd();
	glPopMatrix();

	glPopAttrib();
		
	rtexture->Release();
	
	draw_flow();
}

void print_flow(cstring fname,cstring gname, cstring cname) {
	const int c_xoff=7;
	const int c_yoff=7;	
	const int cap_w=source_w+2*c_xoff;
	const int cap_h=source_h+2*c_yoff;

	unsigned char * data = new unsigned char[3*cap_w*cap_h];


	//rember that the command that calls this has already swapped the buffers.
	glReadBuffer(GL_FRONT);
	glReadPixels(flow_x-c_xoff, flow_y-c_xoff,cap_w,cap_h,GL_RGB,GL_UNSIGNED_BYTE,data);

	fbitmap::write(fname,(sven::rgb *) data,cap_w,cap_h);
	rtexture->Activate();
	
	glReadPixels(0, 0,source_w,source_h,GL_RGB,GL_UNSIGNED_BYTE,data);
	fbitmap::write(gname,(sven::rgb *) data,source_w,source_h);		

	glReadPixels(source_w, 0,source_w,source_h,GL_RGB,GL_UNSIGNED_BYTE,data);
	fbitmap::write(cname,(sven::rgb *) data,source_w,source_h);	
	
	rtexture->Deactivate();

	
	delete [] data;
}

struct RunExample : sven_ui::TemplateCommand {
	void fun(std::vector<Argument> &args, int n) {
		using std::string;
		using sven::qw;
		fbitmap::dims d;
		int i;
		string example="sphere";

		if(n==1) {
			example=args[0].get_string();
			params::sigma=args[1].get_double();
			params::r=args[2].get_double();	
		}

		boost::timer big_t;
		float total_processing_time=0;
		for(i=0;i<=19;i++) {
		
			//calc the optical flow for this frame
			using sven::rgb;
			rgb * data = fbitmap::load3c(
				(fmt("wu/%s/%s.%d.bmp") % example %  example % i).str(),d);
			set_dims(d.w,d.h);
			
			
			boost::timer t;
			send_data((unsigned char *)data);
			update();
			total_processing_time+=t.elapsed();
			
			delete [] data;
			
			glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
			draw();
			SwapBuffers(win_tools::hDC);
			
			//discard the first frame, because it will have garbage data.
			if(i) {
				//print optical flow (along with the gradient and color version of the
				//flow field).
				string gen=(fmt("wu_out/%s/%s-s%.01f-r%d.%d.bmp") 
					% example % example % params::sigma % params::r % i).str();
				string ggen=(fmt("wu_out/%s/g%s-s%.01f-r%d.%d.bmp") 
					% example % example % params::sigma % params::r % i).str();
				string cgen=(fmt("wu_out/%s/c%s-s%.01f-r%d.%d.bmp") 
					% example % example % params::sigma % params::r % i).str();			
					
				print_flow(gen,ggen,cgen);
				sven_ui::report(fmt("generated %s.") % qw(gen));
			}
		}

		sven_ui::report(fmt(
		
		"example run complete.\n"
		"gpu processing time%.02f:\n"
		"total operation time: %.02f") 
			% total_processing_time % big_t.elapsed() );
	}

	RunExample() : TemplateCommand("run_example", 
	"run an example case, optionally takes the example name, sigma and r as arguments", 
	"no args | string double double", "general") {}
};

void register_commands() {
	sven_ui::get_command_manager()->insert(new RunExample);
}

void init() {
	boost::mutex::scoped_lock lock(rt_access_mutex);	

	rtexture = new RenderTexture("float=32 rgb textureRECT", 
		tex_w, tex_h, GL_TEXTURE_RECTANGLE_NV);

	rtexture2=new RenderTexture("float=32 rgb textureRECT", 
		tex_w, tex_h, GL_TEXTURE_RECTANGLE_NV);


	rtexture2->Activate();

    glDisable(GL_DITHER);
    glDisable(GL_DEPTH_TEST);

    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    gluOrtho2D(0, tex_w, 0, tex_h);
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    glViewport(0, 0, tex_w, tex_h);
	
	glClearColor(0,0,0,0);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	rtexture2->Deactivate();

	rtexture->Activate();

    glDisable(GL_DITHER);
    glDisable(GL_DEPTH_TEST);

    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    gluOrtho2D(0, tex_w, 0, tex_h);
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    glViewport(0, 0, tex_w, tex_h);
	
	glClearColor(0,0,0,0);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	rtexture->Deactivate();
	
	
}





void cleanup() {
	delete rtexture;
}

}